#include <stdio.h>
#include <string.h>
#include <math.h>
/* #include <libgen.h> */
#include <assert.h>
#include "proto.h"
#include "headerx.h"
#include "xlate.h"
#include "toolbox.h"
#include "tools.h"
#include "globals.h"

typedef struct {
  LIST *src_flds;
  LIST *tgt_flds;
  LIST *i2efptrs;
  LIST *i2e_misc_info;
} MergeFieldSet;

void KeyList2KeyStr(LIST *kl, char *kstr);
int CopySrcDataToTgtRec(
   char           *src_rec, char *tgt_rec,            /* src & tgt records' raw data pointer */
   int             src_vfd, int   tgt_vfd,            /* src & tgt ".var" descriptors */
   FIELD          *src_fields[], FIELD *tgt_fields[], /* list of src and tgt field descriptors */
   I2EFuncPtr      i2efptr[],                         /* src->tgt conversion functions */
   I2EMiscCnvInfo *i2e_misc_info[],                   /* extra parameters for conversion */
   int nfields                                        /* number of src fields (= tgt fields) */
   );

LIST *KeyStr2KeyFieldList(char *kstr, LABEL *l);
void MakeKeyValStr(DATA d, FIELD *key, char *val_str);
void FreeMergeFieldSet(MergeFieldSet **merge_field_set);
MergeFieldSet *MakeCopySet(char *ifname, LIST *fldl1, char *ofname, LIST *fldl2);
int OutputLabel(char *lbl, char *lbl_name, int rec_len, char *st_key_str,
                char *ed_key_str, char *keys, long num_recs,
                char *fmt_file);
void InitTgtRecByteOffsets(char *rec, LIST *fieldl);
int *FieldOverlap(LABEL *l, int *noverlap);
int *OutOfBoundFields(LIST *fields, int reclen, int **obf, int  *nob);

/***********************************************************************
 ** MAIN ***************************************************************
 ***********************************************************************/
int
main(int ac, char *av[])
{
  char   *f1 = NULL;
  LABEL  *l1;
  FRAGMENT *src_frag;
  char   *f2 = NULL;
  char   *fmt = NULL;
  LABEL  *l2;

  char    st_key_str[512], ed_key_str[512];
  char   *keys = NULL;
  char    keys_buff[1024];

  int     st_key, ed_key;
  DATA    d;

  MergeFieldSet *copy_set;
  int     src_fd, tgt_fd, src_varfd, tgt_varfd;
  int     src_has_vardata, tgt_has_vardata;
  char   *src_data, *tgt_data;
  int     tgt_data_buff_size, tgt_data_offset;
  char   *src_varfname, *tgt_varfname;
  DATA    start_key_data, stop_key_data;
  int     i, j;
  LIST   *keyl;
  FIELD  *primary_key;
  char   *src_rec, *tgt_rec;
  char   *tgt_packed_varfname;
  
  FIELD **src_fields, **tgt_fields;   /* workspace */
  int     n_src_fields, n_tgt_fields;
  I2EMiscCnvInfo **i2emi_list;
  I2EFuncPtr *i2efp_list;
  
  char  *tmp_vfname = NULL; /* temporary target ".var" file name */

  int    *overlap;  /* field overlap matrix for target format sanity */
  int     noverlap; /* count of overlaps */

  int     fspan;    /* holds the maximum field span */
  
  FIELD **fields;   /* working-storage */
  int     nfields;  /* count of fields above */

  int    *obf;      /* working-storage - stores out of bounds bit fields */
  int     nob;      /* count of obf above */

  int     rc;       /* return code */

  int     forced_ignore_overlap = 0; /* User asked to turn off the field-overlap checking */
  
  endian_test();

  /* Verify the correct command line */
  pgm_name = basename(av[0]);
  i = 1;
  while(i < ac){
    if(!strcmp(av[i], "-s")){
      if (++i < ac){
        f1 = av[i];
      }
      else {
        fprintf(stderr, "%s: -s incomplete.\n", pgm_name);
        return -1;
      }
    }
    else if(!strcmp(av[i], "-t")){
      if (++i < ac){
        f2 = av[i];
      }
      else {
        fprintf(stderr, "%s: -t incomplete.\n", pgm_name);
        return -1;
      }
    }
    else if(!strcmp(av[i], "-f")){
      if (++i < ac){
        fmt = av[i];
      }
      else {
        fprintf(stderr, "%s: -f incomplete.\n", pgm_name);
        return -1;
      }
    }
    else if (!strcmp(av[i], "-k")){
      if (++i < ac){
        keys = av[i];
      }
      else {
        fprintf(stderr, "%s: -k incomplete.\n", pgm_name);
        return -1;
      }
    }
    else if (!strcmp(av[i], "-nooverlap")){
      forced_ignore_overlap = 1;
    }
    else {
      fprintf(stderr, "%s: Unrecognized option: %s\n", pgm_name, av[i]);
      return -1;
    }
    i++;
  }

  if ((f1 == NULL) || (f2 == NULL) || (fmt == NULL)){
    fprintf(stderr, "%s Version %s\nUsage:\n %s -s <source.dat> -t <target.dat>\n"
            "          -f <target.fmt> [-k \"'key1'[, 'key2', ..]\"] -nooverlap\n",
				pgm_name, version, pgm_name);
    fprintf(stderr, " ** Note that the first key in the key list will "
            "be used to fill-in the \n    start and stop key values in"
            " the .dat file.\n");
	fprintf(stderr, " -nooverlap disables overlapping fields' checking in the target file.\n");
    return 0;
  }

  /* Verify that the target does not exist already */
  if (access(f2, F_OK) == 0){
    fprintf(stderr, "%s: Target file %s already exists.\n", pgm_name, f2);
    return -2;
  }

  /* Load target file label */
  if ((l2 = LoadFmt(fmt)) == NULL){
    fprintf(stderr, "%s: Fatal! Error while processing %s.\n", pgm_name, fmt);
    return -2;
  }
  /* Note that we cannot check for the existence of a ".var" file associated
     With the target ".dat" file until we open the target label */

  /* verify that the fields are within their parent's boundaries */
  if (OutOfBoundFields(l2->fields, l2->reclen, &obf, &nob) == NULL){
    fprintf(stderr, "%s: Fatal! OutOfBoundFields() failed. ", pgm_name);
    fprintf(stderr, "See previous messages.\n");
  }
  else {
    if (nob > 0){
      fields = (FIELD **)list_data(l2->fields);
      fprintf(stderr, "%s: Fatal! Out-of-bound fields found in %s:\n",
              pgm_name, fmt);
      for (i = 0; i < nob; i++){
        fprintf(stderr, "\t%s\n", fields[obf[i]]->name);
      }
      free(obf);
      return -2;
    }
    /* free(obf); */
  }
  
  if (!forced_ignore_overlap){
    /* generate the field overlap matrix */
    if ((overlap = FieldOverlap(l2, &noverlap)) == NULL){
      fprintf(stderr, "%s: Fatal! FieldOverlap() failed. ", pgm_name);
      fprintf(stderr, "See previous messages.\n");
      return -2;
    }
    else {
      if (noverlap > 0){
        nfields = list_count(l2->fields);
        fields = (FIELD **)list_data(l2->fields);
        
        fprintf(stderr, "%s: Fatal! Following fields overlap in %s.\n", pgm_name, fmt);
        for(i = 0; i < nfields; i++){
          for(j = i+1; j < nfields; j++){
            if (overlap[i*nfields+j]){
              fprintf(stderr, "\t%s <=> %s\n", fields[i]->name, fields[j]->name);
            }
          }
        }
        return -2;
      }
    }
  }
  
  /* get rid of the overlap matrix */
  free(overlap);
  
  tgt_has_vardata = LabelHasByteOffsetFields(l2);
  if (tgt_has_vardata){
    /* If the target ".dat" file is supposed to have a ".var" file associated
       with it, then check for its existence. */
    /* Check if any variations of the ".var" file already exist in the
       target directory. If so, choke. */
    tgt_varfname = GetVarFileName(f2);
    if (tgt_varfname != NULL){
      fprintf(stderr, "%s: Target file %s already exists.\n", pgm_name, tgt_varfname);
      free(tgt_varfname);
      return -2;
    }

    /* Generate a possible ".var" file name for the source ".dat" file. */
    tgt_varfname = MakeVarFileName(f2);
  }


  /* Load source file label */
  if ((l1 = LoadLabel(f1)) == NULL){
    fprintf(stderr, "%s: Error while loading %s.\n", pgm_name, f1);
    return -2;
  }

  /* Build the key string that will go into the target label */
  if (keys != NULL){
    keyl = KeyStr2KeyFieldList(keys, l2);
  }
  else {
    keyl = l1->keys;
  }
  
  if (KeyFieldList2KeyStr(keyl, keys_buff, l2) != list_count(keyl)){
    fprintf(stderr,
            "%s: Warning! Some target keys have been dropped for %s -> %s.\n",
            pgm_name, f1, f2);
  }

  if (strcmp(keys_buff, "") == 0){
    /* If the keys list is empty, abort. */
    fprintf(stderr, "%s: Target keys list is empty for %s.\n", pgm_name, f2);
    return -2;
  }

  /* Assign the first key in the key list as the primary key, this key
     will be used to set the values of START_KEY and STOP_KEY in the
     PDS Label */
  primary_key = ((FIELD **)list_data(keyl))[0];

  /* Get the list of source<->target field pairs to worry about copying */
  if ((copy_set = MakeCopySet(f1, l1->fields, f2, l2->fields)) == NULL){
    fprintf(stderr,
            "%s: Error creating %s -> %s field mapping. See above messages.\n",
            pgm_name, f1, f2);
    return -2;
  }

  /* Load the fragment specific info for the fragment file specified
     by the user */
  src_frag = LoadFragmentWithLabel(f1, l1);
  if (src_frag == NULL){
    fprintf(stderr, "%s: Error loading fragment info from %s.\n", pgm_name, f1);
    return -2;
  }

  /* Load fragment data */
  src_data = (char *)calloc(sizeof(char), src_frag->sbuf.st_size);
  if (src_data == NULL){
    fprintf(stderr, "%s: Not enough memory to load src data file %s.\n", pgm_name, f1);
    return -2;
  }
  
  src_fd = open(f1, O_RDONLY);
  if (src_fd < 0){
    fprintf(stderr, "%s: Could not open source file %s.\n", pgm_name, f1);
    return -2;
  }

  if (read(src_fd, src_data, src_frag->sbuf.st_size) != src_frag->sbuf.st_size){
    fprintf(stderr, "%s: Error reading data %s.\n", pgm_name, f1);
    close(src_fd);
    return -2;
  }

  close(src_fd);

  /* Open the ".var" portion of the source data, if there is one associated with the file */
  src_has_vardata = LabelHasByteOffsetFields(l1);
  if (src_has_vardata){
    if ((src_varfname = GetVarFileName(f1)) == NULL){
      fprintf(stderr, "%s: Cannot find \".var\" file associated with %s.\n", pgm_name, f1);
      return -2;
    }
    
    if ((src_varfd = open(src_varfname, O_RDONLY)) < 0){
      fprintf(stderr, "%s: Error opening %s.\n", pgm_name, src_varfname);
      return -2;
    }
  }

  /* Extract the start- and stop-key values from the data file */
  src_rec = GetRecPtr(src_data, src_frag->offset, l1->reclen, 0);
  start_key_data = ConvertFieldData(src_rec, primary_key);
  MakeKeyValStr(start_key_data, primary_key, st_key_str);
  
  src_rec = GetRecPtr(src_data, src_frag->offset, l1->reclen, src_frag->nrows - 1);
  stop_key_data = ConvertFieldData(src_rec, primary_key);
  MakeKeyValStr(stop_key_data, primary_key, ed_key_str);

  /* Allocate buffer space to accomodate the target file */
  tgt_data_buff_size = src_frag->offset + src_frag->nrows * l2->reclen + 1024;
  assert((tgt_data = (char *)calloc(sizeof(char), tgt_data_buff_size)) != NULL);


  /* Table name is generated from the ".fmt" file name. It is the part of format
     file name just before the ".fmt" - how convenient */
  l2->name = strdup(basename(fmt));
  if (strchr(l2->name, '.')){
    *(strchr(l2->name, '.')) = 0;
  }
  
  /* Output target file's PDS-Header in the target file's buffer space. The header
     generated has its size aligned to the record length of target file. */
  tgt_data_offset = OutputLabel(tgt_data, l2->name, l2->reclen, st_key_str, ed_key_str,
                                keys_buff, src_frag->nrows, fmt);

  /* Adjust the output file buffer space */
  tgt_data_buff_size = tgt_data_offset + src_frag->nrows * l2->reclen;
  assert((tgt_data = (char *)realloc(tgt_data, tgt_data_buff_size)) != NULL);

  /* Open a temporary ".var" file to put the data into */
  tgt_varfd = -1; /* initialize ".var" file descriptor */
  if (tgt_has_vardata){
    tgt_varfd = CreateTmpFile(&tmp_vfname, NULL);
    if (tgt_varfd < 0){
      perror(pgm_name);
      return -2;
    }
  }

  src_fields = (FIELD **)list_data(copy_set->src_flds);
  n_src_fields = list_count(copy_set->src_flds);
  tgt_fields = (FIELD **)list_data(copy_set->tgt_flds);
  n_tgt_fields = list_count(copy_set->tgt_flds);
  i2efp_list = (I2EFuncPtr *)list_data(copy_set->i2efptrs);
  i2emi_list = (I2EMiscCnvInfo **)list_data(copy_set->i2e_misc_info);
  
  /* Copy input records to output file one-by-one */
  for (i = 0; i < src_frag->nrows; i++){
    /* Now fill in the data values being transferred over. Do not
       copy any variable data, we'll move it afterwards. */
    src_rec = GetRecPtr(src_data, src_frag->offset, l1->reclen, i);
    tgt_rec = GetRecPtr(tgt_data, tgt_data_offset, l2->reclen, i);

    /* Initialize the byte offset fields to "-1" indicating that
       there is no variable data associated with the record. */
    InitTgtRecByteOffsets(tgt_rec, l2->fields);
    
    rc = CopySrcDataToTgtRec(src_rec, tgt_rec,        /* <- actual record data ptrs */
                             src_varfd, tgt_varfd,    /* <- ".var" portion descriptors */
                             src_fields, tgt_fields,  /* <- src & tgt field desc (1<->1) */
                             i2efp_list, i2emi_list,  /* <- src->tgt func & func params */
                             n_src_fields);           /* <- #fields to convert #src_fields = #tgt_fields */

    if (!rc){
      /* Error while copying records */
      
      fprintf(stderr, "%s: Error while copying record #%d.\n", pgm_name, i+1);
      
      close(src_varfd);

      if (tgt_has_vardata){
        close(tgt_varfd);
        unlink(tmp_vfname);
      }
      
      return -2;
    }
  }

  /* >>>> Get rid of next line <<<< */
  if (src_has_vardata){
    close(src_varfd);
  }

#if 0
  /* Pack the ".var" portion. This has to happen while the fixed data is still
     in the memory. */
  if (tgt_has_vardata){
    if (src_has_vardata){
      /* If source file had a ".var" portion then create the target ".var" file
         using it. */
      /* We cheat by not creating the new ".var" file along with the fixed records,
         so, at the end of fix record processing, we can simply compact the ".var"
         records based on the target ".dat" file. */
      tgt_packed_varfname = PackVarData(src_varfname, f2, tgt_data_offset,
                                        src_frag->nrows, tgt_data, l2, 1);
      
      /* Copy the packed ".var" file from the temporary location to its correct
         path. */
      copy_file(tgt_packed_varfname, tgt_varfname);

      /* Get rid of the temporary file */
      unlink(tgt_packed_varfname);
    }
    else {
      /* If the source file did not have a ".var" portion, then the target ".var"
         file will be empty. */
      if ((tgt_varfd = open(tgt_varfname, O_RDWR | O_TRUNC | O_CREAT,
                            S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)) < 0){
        fprintf(stderr, "%s: Unable to create %s. ", pgm_name, tgt_varfd); perror("Reason");
        free(tgt_varfname);
        /* Do other cleanup <<<<<<<<<<< */
        return -2;
      }

      close(tgt_varfd);
    }
  }
#else
  if (tgt_has_vardata){
    /* Copy ".var" file from temporary directory to where it is needed. */
    copy_file(tmp_vfname, tgt_varfname);
  }
#endif
  
  /* Copy target ".dat" file's memory based image to the target ".dat" file */
  if ((tgt_fd = open(f2, O_CREAT | O_RDWR | O_TRUNC,
                     S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)) < 0){
    fprintf(stderr, "%s: Could not create target file %s. ", pgm_name, f2); perror("Reason");
    unlink(tgt_varfname);
    return -2;
  }

  if (write(tgt_fd, tgt_data, tgt_data_buff_size) != tgt_data_buff_size){
    fprintf(stderr, "%s: Unable to write to %s. ", pgm_name, f2); perror("Reason");
    unlink(tgt_varfname);
    return -2;
  }

  close(tgt_fd);
  
  /* Free up the resources */
  free(src_data);
  free(tgt_data);
  FreeFragment(src_frag);

  /*
  FreeLabel(l1);
  FreeLabel(l2);
  */
  
  if (src_has_vardata){
    free(src_varfname);
  }

  if (tgt_has_vardata){
    free(tgt_varfname);
  }

  /* Return success */
  return 0;
}

/*
** Returns a list of indices of fields that are out of bounds
** of their respective parents. For bit-fields, these are their
** parent bit-strings, whereas, for regular fields, it is the
** record-length.
**
** The indices are returned in "*obf" and the count is returned
** in "*nob".
**
** It is the caller's responsibility to free the memory malloc'ed
** to "*obf".
*/
int *
OutOfBoundFields(
    LIST *fields,  /* input: list of fields */
    int   reclen,  /* record length of file to which "fields" belong to */
    int **obf,     /* output: indices of out of bound fields */
    int  *nob      /* number of out-of-bound fields found */
    )
{
  FIELD **f;
  int   i, n;

  f = (FIELD **)list_data(fields);
  n = list_count(fields);
  
  *nob = 0;
  *obf = (int *)calloc(sizeof(int), n);
  if (*obf == NULL){
    fprintf(stderr, "Vcol: Mem allocation failure.\n");
    return NULL;
  }
  
  for(i = 0; i < n; i++){
    if (f[i]->bitfield && strchr(f[i]->name, ':')){
      /* bit-fields */
      if ((f[i]->bitfield->start_bit + f[i]->bitfield->bits - 1) >
          BYTES_TO_BITS(f[i]->size)){
        (*obf)[(*nob)++] = i;
      }
    }
    else {
      /* non- bit-fields */
      if ((f[i]->start + f[i]->size * (f[i]->dimension? f[i]->dimension: 1) - 1) > reclen){
        (*obf)[(*nob)++] = i;
      }
    }
  }

  /* fit output array to size */
  *obf = (int *)realloc(*obf, sizeof(int) * (*nob) + 1);
  
  return *obf;
}


/*
** Returns "l->nfields" x "l->nfields" matrix with a non-zero
** indicating overlap. The matrix is single-dimensioned. It is
** allocated using malloc() and it is the caller's responsibility
** to free it after its use. The "*noverlap" value gives the
** actual overlaps detected, such that, the f[i] <-> f[i] overlap
** and is not counted and f[i] <-> f[j] & f[j] <-> f[i] overlap
** is not counted twice.
**
** CAUTION: It will ignore bit-field <-> bit-string overlap
** within an MSB_BIT_STRING.
**
** On error NULL is returned.
*/
int *
FieldOverlap(LABEL *l, int *noverlap)
{
  int *overlap;
  FIELD **f;
  int   nfields;
  int i, j;
  char *p, *q;

  f = (FIELD **)list_data(l->fields);
  nfields = list_count(l->fields);

  *noverlap = 0;
  
  /*
  ** Allocate an "nfields" x "nfields" overlap indicating matrix
  */
  overlap = (int *)calloc(sizeof(int), nfields * nfields);
  if (overlap == NULL){
    fprintf(stderr, "%s: Mem allocation failed.\n", pgm_name);
    return NULL;
  }

  for(i = 0; i < nfields; i++){
    overlap[i*nfields+i] = 1;

    /* check if the field has the bit-field marker in it */
    p = strchr(f[i]->name, ':');
    
    for(j = i+1; j < nfields; j++){
      
      /* check if the field has the bit-field marker in it */
      q = strchr(f[j]->name, ':');

      /*
      ** NOTE: For a field to be a bit-field, it's "bitfield"
      ** component must be non-NULL and it must have the
      ** bit-field marker in it.
      **
      ** NOTE: bit-strings should be compared as regular
      ** fields.
      */

      if ((f[i]->bitfield && p) && (f[j]->bitfield && q)){
        /*
        ** If both fields are bit-fields then we want to do
        ** overlap test. However, they must belong to the
        ** same parent bit-string field. This can be verified
        ** by comparing the field-name portion before the ":".
        */
        if ((int)(p - f[i]->name) == (int)(q - f[j]->name) &&
            strncasecmp(f[i]->name, f[j]->name, (int)(p - f[i]->name)) == 0){
          /*
          ** At this point we've established that both the
          ** bit-fields belong to the same parent bit-string
          ** field.
          */

          if ((f[i]->bitfield->start_bit + f[i]->bitfield->bits) <=
              f[j]->bitfield->start_bit){
            /* f[i] is totally outside f[j] to its left */
            overlap[i*nfields+j] = overlap[j*nfields+i] = 0;
          }
          else if (f[i]->bitfield->start_bit >=
                   (f[j]->bitfield->start_bit + f[j]->bitfield->bits)){
            /* f[i] is totally outside f[j] to its left */
            overlap[i*nfields+j] = overlap[j*nfields+i] = 0;
          }
          else {
            overlap[i*nfields+j] = overlap[j*nfields+i] = 1;
            (*noverlap) ++;
          }
        }
        else {
          /*
          ** The two bit-fields don't belong to the same
          ** bit-string. So, by default, they don't overlap,
          ** unless their parent's overlap. It is a big
          ** assumption based on the fact that the bit-fields
          ** don't go beyond the bounds of their parent
          ** bit-string fields.
          */
          overlap[i*nfields+j] = overlap[j*nfields+i] = 0;
        }
      }
      else if ((f[i]->bitfield && p) || (f[j]->bitfield && q)){
        /*
        ** Don't compare bit-fields to bit-strings or regular
        ** fields. Assume that they don't overlap!
        */
        overlap[i*nfields+j] = overlap[j*nfields+i] = 0;
      }
      else if ((f[i]->start + f[i]->size * (f[i]->dimension? f[i]->dimension: 1)) <= f[j]->start){
        /* f[i] is totally outside f[j] to the left */
        overlap[i*nfields+j] = overlap[j*nfields+i] = 0;
      }
      else if (f[i]->start >= (f[j]->start + f[j]->size * (f[i]->dimension? f[i]->dimension: 1))){
        /* f[i] is totally outside f[j] to the right */
        overlap[i*nfields+j] = overlap[j*nfields+i] = 0;
      }
      else {
        /* f[i] & f[j] overlap */
        overlap[i*nfields+j] = overlap[j*nfields+i] = 1;
        (*noverlap) ++;
      }
    }
  }

  return overlap;
}



/**
 ** Extracted from: /tes/src/vanilla_stuff/fftfix.c
 ** ***************************************************************************
 ** Converts the given vector of doubles into a vector of shorts and exponent.
 **
 ** Args:
 **     double *f:  Given vector of doubles
 **     int len:    Length of vector
 **     short *s:   Resultant vector of shorts. Caller should mem_alloc
 **     short *expo:    Exponent for vector 's'
 **/
void
IntoQ15(double *f, int len, short *s, short *expo)
{
    int i;
    short e = 0;
    double max, v, factor, ln2;

    max = f[0];
    ln2 = log(2);

    for (i = 0; i < len; i++) {
        v = fabs(f[i]);
        if (max < v)
            max = v;
    }
    /**
    ** All zeros.  Cope
    **/
    if (max == 0) {
        for (i = 0; i < len; i++)
            s[i] = 0;
        *expo = 0;
        return;
    }
    /**
    ** Scale by factors of 2 until we are as close to 1.0 as we can get.
    **/
    if (max > 1.0) {
        while (max > 1.0) {
            max = max / 2.0;
            e++;
        }
    } else if (max < 1.0) {
        while (max < 1.0) {
            max = max * 2.0;
            e--;
        }
    /**
        ** Just over 1.0, go back one step.
        **/
        e++;
    }
    /**
    ** Adjust everything to the scale factor
    **/
    if ((15 - e) >= 0)
        factor = ceil(exp((15 - e) * ln2));
    else
        factor = 1.0 / ceil(exp(-(15 - e)) * ln2);

    for (i = 0; i < len; i++)
        s[i] = f[i] * factor;

    *expo = e;
    return;
}



/*
** Convert the raw field data into a real number (double).
*/

double
ConvertNumData(char *field_data, FIELD *f)
{
  DATA   d;
  double r = 0.0;

  d = ConvertData(field_data, f);
  
  switch(f->iformat){
  case INT:
    r = (double)d.i;
    break;
  case UINT:
    r = (double)d.ui;
    break;
  case REAL:
    r = d.r;
    break;
  default:
    fprintf(stderr, "%s Inappropriate call to ConvertNumData()\n", pgm_name);
    abort();
  }

  /* Apply offset and scaling, where applicable. */
  if (f->scale != 0){ r *= f->scale; }
  if (f->offset != 0){ r += f->offset; }

  return r;
}



/*
** Following function takes a pointer to a record's raw data and a pointer
** to its field description (FIELD *), extracts the data from the record,
** converts it into Q15 data (composed of a common exponent and (short)
** values), envelopes this Q15 data into the data-byte count (VAX Variable
** Record's header and trailer) and returns it.
**
** The output array of shorts is allocated using malloc() and it is the
** caller's responsibility to free the memory after use.
**
** The caller should get a count of words by using the first short value
** i.e., the VAX Variable Record's header.
*/

short *
NumArray2Q15VaxVar(
  char  *src_rec,
  FIELD *sf
)
{
  double *vals;
  char   *src_field;
  short  *q15data;
  int     i;

  vals = (double *)calloc(sf->dimension, sizeof(double));
  if (vals == NULL){
    fprintf(stderr, "%s: Mem allocation failed.\n", pgm_name);
    return NULL; /* failure */
  }

  /*
  ** Allocate enough space for the full VAX variable record
  ** i.e., the header & trailer words, the data and its
  ** common exponent.
  */
  q15data = (short *)calloc(sf->dimension+3, sizeof(short));
  if (q15data == NULL){
    fprintf(stderr, "%s: Mem allocation failed.\n", pgm_name);
    free(vals);
    return NULL; /* failure */
  }

  /* Extract and convert the input values to floating point */
  for(i = 0; i < sf->dimension; i++){
    src_field = GetFieldPtr(src_rec, sf, i);
    vals[i] = ConvertNumData(src_field, sf);
  }

  /* Convert the values into Q15 */
  IntoQ15(vals, sf->dimension, /* vals: */ &q15data[2], /* exp: */ &q15data[1]);
  
  /* Fill in VAX Var data header and trailer, which should contain "vals+exp"'s length */
  q15data[0] = q15data[sf->dimension + 2] = sizeof(short) * (sf->dimension + 1);

  /* Don't need the float values any more, get rid of them. */
  free(vals);
  
  /* Pass the Q15 data back. */
  return q15data; /* success */
}


/*--------------------------------*/
/** Src: Fixed-length Array      **/
/** Tgt: Variable Array          **/
/*--------------------------------*/
#define IS_EXCEPTION_CASE(f1,f2) \
   (((f1)->iformat == INT || (f1)->iformat == UINT || (f1)->iformat == REAL) && \
    (f1)->dimension > 0 && (f2)->vardata && (f2)->vardata->type == Q15)


int
CopySrcDataToTgtRec(
   char           *src_rec, char *tgt_rec,            /* src & tgt records' raw data pointer */
   int             src_vfd, int   tgt_vfd,            /* src & tgt ".var" descriptors */
   FIELD          *src_fields[], FIELD *tgt_fields[], /* list of src and tgt field descriptors */
   I2EFuncPtr      i2efptr[],                         /* src->tgt conversion functions */
   I2EMiscCnvInfo *i2e_misc_info[],                   /* extra parameters for conversion */
   int nfields                                        /* number of src fields (= tgt fields) */
   )
{
  int    i, j;
  int    min_of_maxs;
  char  *src_fld;
  char  *tgt_fld;
  DATA   src_data;
  short *q15data;
  int    nels;
  int    new_offset = -1;
  int    odsz;  /* output data size (convenience variable) */

  char  *var_rec = NULL; /* variable data buffer and its allocated size */
  int    var_rec_buff_sz = 0;

  for (i = 0; i < nfields; i++){
    if (IS_EXCEPTION_CASE(src_fields[i], tgt_fields[i])){
      /*-------------------------------------------*/
      /* Fixed-length field -> Variable data field */
      /*-------------------------------------------*/

      /* Extract variable data, convert to Q15 and package as Vax Variable data */
      q15data = NumArray2Q15VaxVar(src_rec, src_fields[i]);
      if (q15data == NULL){
        fprintf(stderr, "%s: Mem allocation error.\n", pgm_name);
        
        if (var_rec){ free(var_rec); }
        
        return 0;
      }

      /* Get variable data's address from the ".var" file. */
      src_data.i = lseek(tgt_vfd, 0, SEEK_CUR);
      
      /* Save the VAX Variable data in ".var" */
      /* odsz = sizeof(short)*2 + q15data[0]; */
      nels = src_fields[i]->dimension+1; /* data + exp */
      odsz = VardataBuffSize(tgt_fields[i]->vardata, nels);
      if (write(tgt_vfd, (char *)q15data, odsz) != odsz){
        perror(pgm_name);
        
        if (var_rec){ free(var_rec); }
        free(q15data);
        
        return 0;
      }
      
      /* Save the address of variable data in the ".dat" file. */
      tgt_fld = GetFieldPtr(tgt_rec, tgt_fields[i], 0);
      (*i2efptr[i])(src_data, tgt_fld, tgt_fields[i]->size, NULL, i2e_misc_info[i]);
      
      free(q15data);
    }
    else if (src_fields[i]->vardata != NULL && tgt_fields[i]->vardata != NULL){
      /*--------------------------------------------*/
      /* Variable data field -> Variable data field */
      /*--------------------------------------------*/

      /* Read variable data offset from the source file. */
      src_fld = GetFieldPtr(src_rec, src_fields[i], 0);
      src_data = ConvertData(src_fld, src_fields[i]);
      
      /* Read the variable record */
      nels = ReadVardataRec(src_vfd, src_data.i, src_fields[i], &var_rec, &var_rec_buff_sz);

      /* Get variable data's address from the ".var" file. */
      src_data.i = lseek(tgt_vfd, 0, SEEK_CUR);
      
      /* Save the Variable data in ".var" file. */
      odsz = VardataBuffSize(tgt_fields[i]->vardata, nels);
      if (write(tgt_vfd, var_rec, odsz) != odsz){
        perror(pgm_name);

        if (var_rec){ free(var_rec); }

        return 0;
      }

      /* Save the address of variable data in the ".dat" file. */
      tgt_fld = GetFieldPtr(tgt_rec, tgt_fields[i], 0);
      (*i2efptr[i])(src_data, tgt_fld, tgt_fields[i]->size, NULL, i2e_misc_info[i]);
    }
    else if (src_fields[i]->dimension > 0){
      /*----------------*/
      /* Array -> Array */
      /*----------------*/
      
      min_of_maxs = src_fields[i]->dimension < tgt_fields[i]->dimension ?
        src_fields[i]->dimension : tgt_fields[i]->dimension;

      for (j = 0; j < min_of_maxs; j++){
        src_fld = GetFieldPtr(src_rec, src_fields[i], j);
        src_data = ConvertData(src_fld, src_fields[i]);
        
        tgt_fld = GetFieldPtr(tgt_rec, tgt_fields[i], j);
        (*i2efptr[i])(src_data, tgt_fld, tgt_fields[i]->size, NULL, i2e_misc_info[i]);
      }
    }
    else {
      /*------------------------------*/
      /* Atomic field -> Atomic field */
      /*------------------------------*/
      
      src_fld = GetFieldPtr(src_rec, src_fields[i], 0);
      src_data = ConvertData(src_fld, src_fields[i]);
        
      tgt_fld = GetFieldPtr(tgt_rec, tgt_fields[i], 0);
      (*i2efptr[i])(src_data, tgt_fld, tgt_fields[i]->size, NULL, i2e_misc_info[i]);
    }
  }

  if (var_rec){ free(var_rec); }
  
  return 1; /* success */
}

/* >>>> REQ'D <<<< */
LIST *
KeyStr2KeyFieldList(char *kstr, LABEL *label)
{
  char *p, *q, *r;
  LIST *l;
  FIELD *f;

  if ((l = new_list()) != NULL){
    for (q = kstr; p = strtok(q, ","); q = NULL){
      /* strip leading spaces, tabs, and single quotes */
      while(*p && strchr(" \t\'", *p)){
        p++;
      }

      /* strip following spaces, tabs, and single quotes */
      while(*p && strchr(" \t\'", p[strlen(p)-1])){
        p[strlen(p)-1] = 0;
      }

      /* Add element to the list if non-empty */
      if (strcmp(p, "") != 0){
        if ((f = FindFieldInLabel(p, label)) != NULL){
          list_add(l, f);
        }
        else {
          fprintf(stderr, "%s: Cannot find %s.%s\n", p, label->name);
          list_free(l);
          return NULL;
        }
      }
    }
  }
  
  return l;
}


void
MakeKeyValStr(DATA d, FIELD *key, char *val_str)
{
  switch(key->iformat){
  case INT:
    sprintf(val_str, "%d", d.i);
    break;

  case UINT:
    sprintf(val_str, "%u", d.ui);
    break;
    
  case REAL:
    sprintf(val_str, "%-.2f", d.r);
    break;

  case STRING:
    sprintf(val_str, "%-*.*s", key->size, key->size, d.str);
    break;
  }
} 



/* >>>> REQ'D <<<< */
/***********************************************************************
 ** KeyList2KeyStr() - Converts a Key LIST to a keys string. This can **
 **                    then directly be put into a table fragment in  **
 **                    its header.                                    **
 ***********************************************************************/
int
KeyFieldList2KeyStr(
  LIST  *kl,    /* Source key list */
  char  *kstr,  /* Output key string */
  LABEL *label  /* check if src-key exists in specified label */
)
{
  int     i, kCt;
  FIELD **k;
  char    buff[50];
  FIELD  *f;
  int     key_ct_keystr;

  strcpy(kstr, "");
  if (kl == NULL)
    return;

  k = (FIELD **)kl->data(kl);
  kCt = kl->count(kl);
  key_ct_keystr = 0;
  
  for (i = 0; i < kCt; i++){
    f = FindFieldInLabel(k[i]->name, label);
    if (f != NULL){
      /* Put the source key in the target list only if it also exists
         there. */
      if (strcmp(kstr, "") != 0){
        strcat(kstr, ",");
      }
      sprintf(&kstr[strlen(kstr)], "\'%s\'", f->name);
      key_ct_keystr ++;
    }
  }

  return key_ct_keystr;
}



void
FreeMergeFieldSet(MergeFieldSet **merge_field_set)
{
  int i, ni2e_info;
  I2EMiscCnvInfo **i2e_info;
  
  
  if ((*merge_field_set)->src_flds){ list_free((*merge_field_set)->src_flds); }
  if ((*merge_field_set)->tgt_flds){ list_free((*merge_field_set)->tgt_flds); }
  if ((*merge_field_set)->i2efptrs){ list_free((*merge_field_set)->i2efptrs); }
  if ((*merge_field_set)->i2e_misc_info){
    ni2e_info = list_count((*merge_field_set)->i2e_misc_info);
    i2e_info = (I2EMiscCnvInfo **)list_data((*merge_field_set)->i2e_misc_info);

    for (i = 0; i < ni2e_info; i++){
      if (i2e_info[i] != NULL){ free(i2e_info[i]); }
    }
    list_free((*merge_field_set)->i2e_misc_info);
  }

  free(*merge_field_set); *merge_field_set = NULL;
}



MergeFieldSet *
MakeCopySet(char *ifname, LIST *fldl1, char *ofname, LIST *fldl2)
{
  FIELD  **flds1;
  FIELD  **flds2;
  int      nflds1, nflds2;
  int      i, j;
  int      found;
  MergeFieldSet *copy_set;
  I2EMiscCnvInfo *misc_info;
  I2EFuncPtr i2efptr;
  IFORMAT iformat;
  EFORMAT eformat;

  assert((copy_set = (MergeFieldSet *)calloc(sizeof(MergeFieldSet), 1)) != NULL);
  assert((copy_set->src_flds = new_list()) != NULL);
  assert((copy_set->tgt_flds = new_list()) != NULL);
  assert((copy_set->i2efptrs = new_list()) != NULL);
  assert((copy_set->i2e_misc_info = new_list()) != NULL);

  flds1 = (FIELD **)list_data(fldl1);
  nflds1 = list_count(fldl1);
  flds2 = (FIELD **)list_data(fldl2);
  nflds2 = list_count(fldl2);

  for (i = 0; i < nflds1; i++){
    found = 0;
    for (j = 0; j < nflds2; j++){
      if (strcasecmp(flds1[i]->name, flds2[j]->name) == 0){
        found = 1;
        break;
      }
    }

    if (found){
      /*
      ** If a field name match is found, check to see if the fields
      ** are acutally compatible.
      **
      ** Fields are incompatible if:
      ** - They have different dimensions
      **   (with the exception that a integer/real array
      **   field can be converted into a variable data field)
      */
      if (IS_EXCEPTION_CASE(flds1[i], flds2[j])){
        /* This is the exception case mentioned above. */
      }
      else {
        if (((flds1[i]->dimension > 0) && (flds2[j]->dimension <= 0)) ||
            ((flds1[i]->dimension <= 0) && (flds2[j]->dimension > 0))){
          fprintf(stderr, "%s: Translation from %s(%s) -> %s(%s) is not supported.\n",
                  pgm_name, ifname, flds1[i]->name, ofname, flds2[j]->name);
          FreeMergeFieldSet(&copy_set);
          return NULL;
        }
      }
      
      list_add(copy_set->src_flds, flds1[i]);
      list_add(copy_set->tgt_flds, flds2[j]);

      /* Fill in the miscellaneous info to enable us to do the conversion from the internal
         representation of (one element of) the source field data to the external
         representation of (one element of) the target field. */
      if (flds2[j]->bitfield != NULL){
        assert((misc_info = (I2EMiscCnvInfo *)calloc(sizeof(I2EMiscCnvInfo), 1)) != NULL);
        misc_info->bitfield.shift = flds2[j]->bitfield->shifts;
        misc_info->bitfield.mask = flds2[j]->bitfield->mask;
      }
      else if (flds2[j]->eformat == CHARACTER){
        assert((misc_info = (I2EMiscCnvInfo *)calloc(sizeof(I2EMiscCnvInfo), 1)) != NULL);
        misc_info->string.input_field_size = flds1[i]->size;
      }
      else {
        misc_info = NULL;
      }
      list_add(copy_set->i2e_misc_info, misc_info);

      if (IS_EXCEPTION_CASE(flds1[i], flds2[j])){
        /*
        ** This function does not fit into the normal logic of Vcol, that's
        ** it is being kept separate.
        */

        i2efptr = I2E[INT][BYTE_OFFSET];
      }
      else {
        /* Fill in the function that is required to do the copy */
        iformat = flds1[i]->iformat;
        eformat = flds2[j]->eformat;
        
        if (flds2[j]->eformat == MSB_BIT_FIELD && flds2[j]->bitfield == NULL){
          eformat = MSB_INTEGER;
        }
        
        if ((iformat > 0 && iformat <= MAX_IFORMATS) &&
            (eformat > 0 && eformat <= MAX_EFORMATS)){
          i2efptr = I2E[iformat][eformat];
        }
        else {
          i2efptr = NULL;
        }
      }
      
      if (i2efptr == NULL){
        fprintf(stderr, "%s: Translation from %s(%s) -> %s(%s) is not supported.\n",
                pgm_name, ifname, flds1[i]->name, ofname, flds2[j]->name);
        FreeMergeFieldSet(&copy_set);
        return NULL;
      }
      
      list_add(copy_set->i2efptrs, i2efptr);
    }
  }

  return copy_set;
}


/* >>>> REQ'D <<<< */
/***********************************************************************
 ** OutputLabel() - Stick in a PDS label in the output file as the   **
 **                  very first thing.                                **
 ***********************************************************************/
int
OutputLabel(char *lbl, char *lbl_name, int rec_len, char *st_key_str,
            char *ed_key_str, char *keys, long num_recs,
            char *fmt_file)
{
  int  tbl_start;
  int  ct;
  char lbl_skel[] = 
    "RECORD_TYPE              = FIXED_LENGTH\n"
    "RECORD_BYTES             = %-10d\n"
    "DATA_SET_ID              = \"VANILLA-%s\"\n"
    "\n"
    "^TABLE                   = %-10d\n"
    "\n"
    "OBJECT                   = TABLE\n"
    "    NAME                 = %s\n"
    "    INTERCHANGE_FORMAT   = BINARY\n"
    "    START_KEY            = %-20s\n"
    "    STOP_KEY             = %-20s\n"
    "    KEYS                 = {%s}\n"
    "    ROWS                 = %-20d\n"
    "\n"
    "    ^STRUCTURE           = \"%s\"\n"
    "END_OBJECT               = TABLE\n"
    "END\n";
  char *p;

  /* We don't know the ^TABLE = value right now, so, we just assume
     it as zero and pad enough spaces to hold the largest table start
     value. */
  sprintf(lbl, lbl_skel, rec_len, lbl_name, 0, lbl_name, st_key_str,
          ed_key_str, keys, num_recs, basename(fmt_file));

  /* Next we determine the label size depending upon the label ("lbl")
     that we just generated in the previous step. The ^TABLE or the
     table start value is an integral number >= (label length / rec length) */
  tbl_start = strlen(lbl) / rec_len;
  if (strlen(lbl) % rec_len){
    tbl_start ++;
  }

  /* Install the ^TABLE value back into the label */
  p = strstr(lbl, "^TABLE");
  p = strstr(p, "= ");
  p += strlen("= ");
  sprintf(p, "%d", tbl_start+1);
  p[strlen(p)] = ' '; /* get rid of the string terminator '\0' */

  /* Pad spaces at the end of label to make it an integral multiple of
     the record length. Actually, the label size must be:
     record length * table start.  */
  while(strlen(lbl) < (tbl_start * rec_len) - 1){
    strcat(lbl, " ");
  }

  /* For neatness in the data file, put a carriage return as the last
     label character instead of the space. This is done only if the
     label length was found earlier as not being a multiple of the
     record length */
  if (strlen(lbl) < (tbl_start * rec_len)){
    strcat(lbl, "\n");
  }

  return tbl_start * rec_len; /* successful installation of the label */
}

/* >>>> REQ'D <<<< */
/***********************************************************************
 ** InitTgtRecByteOffsets() - Initalize the output record. We are basically   **
 **             concerned with the BYTE_OFFSET or the variable data   **
 **             offset fields, which must be initialized to -1L until **
 **             and unless there is variable data.                    **
 ***********************************************************************/
void
InitTgtRecByteOffsets(char *rec, LIST *fieldl)
{
  FIELD **fields = (FIELD **)fieldl->data(fieldl);
  int     nfields = fieldl->count(fieldl);
  int     i;
  int     no_vardata = BYTE_OFFSET_NO_VAR_DATA;
  
  for (i = 0; i < nfields; i++){
    if (fields[i]->vardata != NULL){
      /* If field is a byte offset field, then initialize it with
         the byte offset sentinal, i.e. -1 */
      memcpy(GetFieldPtr(rec, fields[i], 0), (void *)&no_vardata, sizeof(no_vardata));
    }
  }
}




