/**
 ** Module: copy_file.c
 **
 ** Copy the source file to the target file.
 ** Returns 0 on failure, or 1 on success.
 **
 **/

#include "globals.h"
#include <stdio.h>

#define COPY_BUFF_SZ 2048


int
copy_file(char *source_file, char *target_file)
{
  FILE *source_fp, *target_fp;
  char buff[COPY_BUFF_SZ];
  int  bytes_read, bytes_written;
  
  VERBOSE(fprintf(stderr, "%s: Copying %s -> %s...\n",
				pgm_name, source_file, target_file));

  if ((source_fp = fopen(source_file, "rb")) == NULL){
    fprintf(stderr, "%s: read-open of file %s failed for copy operation.\n",
            pgm_name, source_file);
    return 0;
  }
  
  if ((target_fp = fopen(target_file, "wb")) == NULL){
    fprintf(stderr, "%s: write-open of file %s failed for copy operation.\n",
            pgm_name, target_file);
    fclose(target_fp);
    return 0;
  }
  
  do{
    bytes_read = fread(buff, sizeof(char), COPY_BUFF_SZ, source_fp);
    if (ferror(source_fp)){
      fprintf(stderr, "%s: read failed for copy from %s to %s at offset %ld.\n",
              pgm_name, source_file, target_file, ftell(target_fp));
      fclose(source_fp);
      fclose(target_fp);
      return 0;
    }
    
	 if (bytes_read > 0){
		 if ((bytes_written = fwrite(buff, sizeof(char), bytes_read, target_fp)) != bytes_read){
			fprintf(stderr, "%s: write failed for copy from %s to %s at offset %ld.\n",
					  pgm_name, source_file, target_file, ftell(source_fp));
			fclose(source_fp);
			fclose(target_fp);
			return 0;
		 }
	 }
  }while (bytes_read == COPY_BUFF_SZ);
  
  fclose(source_fp);
  fclose(target_fp);
  
  return 1;
}
