#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

/**
 ** Function: CreateTmpFile()
 **
 ** Input:
 **   tmp_fn_pfx - Temporary file-name prefix (a character string)
 **
 ** Output:
 **   tmp_fname - Pointer to a blank/unallocated char* buffer where the
 **               name of temporary file is returned. It is allocated
 **               using malloc()
 **
 ** Returns:
 **   File descriptor of the temporary file created and opened
 **
 ** Notes:
 **   First preference in file creation is given to /tmp directory
 **   in order to create the file. The file is always created in
 **   write-only mode.
 **
 ** >>>>> Should give the option of specifying creation mode
 **       and permissions.
 **/
int
CreateTmpFile(char **tmp_fname, char *tmp_fn_pfx)
{
  int tmp_fd = -1;
  
  if ((*tmp_fname = tempnam("/tmp", tmp_fn_pfx)) != NULL){
    /* Open the temporary file */
    if ((tmp_fd = open(*tmp_fname, O_CREAT | O_RDWR, S_IRWXU | S_IRWXG)) < 0){
      free(*tmp_fname);
      *tmp_fname = NULL;
    }
  }
  
  return tmp_fd;
}

int
exists(const char *fname)
{
  struct stat stat_buf;
  
  if (stat(fname, &stat_buf) == 0){
    return 1;
  }

  return 0;
}
